# Copyright (c) 2007-2008 Forest Bond.
# This file is part of the pytagsfs software package.
#
# pytagsfs is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import sys, traceback, locale, logging
from logging import DEBUG, INFO, WARNING, ERROR, CRITICAL

try:
    from functools import wraps
except ImportError:
    from sclapp.legacy_support import wraps

from pytagsfs.logfile import VirtualLogFile

class VirtualLogFileStream(object):
    def write(self, s):
        VirtualLogFile.log_write(s)

    def flush(self):
        pass

logger = logging.getLogger()
logger.addHandler(logging.StreamHandler(VirtualLogFileStream()))

def set_log_level(lvl):
    logger.setLevel(lvl)

def set_logsize(size):
    VirtualLogFile.set_max_length(size)

stderr_enabled = False
def enable_stderr():
    global stderr_enabled
    if not stderr_enabled:
        logger.addHandler(logging.StreamHandler())
        stderr_enabled = True

def log_debug(s):
    logger.debug(s)

def log_info(s):
    logger.info(s)

def log_warning(s):
    logger.warning(s)

def log_error(s):
    logger.error(s)

def log_critical(s):
    logger.critical(s)

def log_traceback():
    log_critical(traceback.format_exc())

def ignore_exceptions(*ignored_exceptions):
    def decorator(fn):
        @wraps(fn)
        def new_fn(*args, **kwargs):
            try:
                return fn(*args, **kwargs)
            except ignored_exceptions, e:
                pass
        return new_fn
    return decorator

def exceptionally_logged(ignore):
    decorate_immediately = False
    if hasattr(ignore, '__call__'):
        decorate_immediately = True
        fn = ignore
        ignore = ()

    def decorator(fn):
        @wraps(fn)
        def new_fn(*args, **kwargs):
            try:
                return fn(*args, **kwargs)
            except ignore:
                raise
            except Exception:
                log_traceback()
                raise
        return new_fn

    if decorate_immediately:
        return decorator(fn)
    return decorator

def logged(fn):
    @wraps(fn)
    def new_fn(*args, **kwargs):
        log_debug('Entering %s' % repr(fn))
        try:
            retval = fn(*args, **kwargs)
        finally:
            log_debug('Leaving %s' % repr(fn))
        return retval
    return new_fn

def verbosely_logged(fn):
    @wraps(fn)
    def new_fn(*args, **kwargs):
        log_debug('Entering %s: args=%s, kwargs=%s' % (
          repr(fn), repr(args), repr(kwargs)))
        try:
            retval = fn(*args, **kwargs)
        except Exception, e:
            log_debug('Exception raised in %s: %s' % (repr(fn), repr(e)))
            raise
        log_debug('Leaving %s: returning %s' % (repr(fn), repr(retval)))
        return retval
    return new_fn
