# Copyright (c) 2007-2008 Forest Bond.
# This file is part of the pytagsfs software package.
#
# pytagsfs is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import sys, os, errno, fcntl, re, locale, shutil

try:
    from functools import wraps
except ImportError:
    from sclapp.legacy_support import wraps

from optparse import OptParseError, SUPPRESS_HELP
from threading import Lock, RLock

import fuse
from fuse import Fuse, Stat, FuseOptParse

from pytagsfs.pathparser import PathParser
from pytagsfs.sourcetree import SourceTree
from pytagsfs.metaparser import (
  DelegateMultiMetaParser,
  FilenameMetaParser,
  ParentDirMetaParser,
  ExtensionMetaParser,
  MutagenFileMetaParser,
)
from pytagsfs.exceptions import (
  PathNotFound,
  NotADirectory,
  PathExists,
  InvalidArgument,
  DirectoryNotEmpty,
  MissingDependency,
)
from pytagsfs.pathpropcache import PathPropCache
from pytagsfs.debug import (
  logged,
  verbosely_logged,
  exceptionally_logged,
  log_debug,
  log_info,
  log_warning,
  log_error,
  set_log_level,
  DEBUG,
  INFO,
  WARNING,
  ERROR,
  CRITICAL,
  enable_stderr,
  set_logsize,
)
from pytagsfs.util import (
  get_obj_by_dotted_name,
  nonatomic,
  attr_named,
  stat_result_to_fuse_stat,
)
from pytagsfs.logfile import VirtualLogFile
from pytagsfs import __version__ as version

fuse.fuse_python_api = (0, 2)

DEFAULT_IOCHARSET = 'utf-8'
DEFAULT_SOURCE_IOCHARSET = 'utf-8'
DEFAULT_VERBOSITY = 'warning'

def append_filter(option, opt_str, value, parser, *args, **kwargs):
    if not hasattr(parser.values, 'filters'):
        parser.values.filters = [ ]
    if 'src' in opt_str:
        which = 'src'
    else:
        which = 'dst'
    parser.values.filters.append((which, value))

def return_errno(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except PathNotFound:
            return -errno.ENOENT
        except NotADirectory:
            return -errno.ENOTDIR
        except PathExists:
            return -errno.EEXIST
        except DirectoryNotEmpty:
            return -errno.ENOTEMPTY
        except InvalidArgument:
            return -errno.EINVAL
        except NotImplementedError:
            return -errno.ENOSYS
        except (IOError, OSError), e:
            if hasattr(e, 'errno') and e.errno > 0:
                return -e.errno
            raise
    return wrapper

def operation_on_one_real_path(real_op):
    @return_errno
    @wraps(real_op)
    def wrapper(self, fake_path, *args, **kwargs):
        real_path = self.encode_real_path(self.get_real_path(
          self.decode_fake_path(fake_path)))
        return real_op(real_path, *args, **kwargs)
    return wrapper

def operation_on_two_real_paths(real_op):
    @return_errno
    @wraps(real_op)
    def wrapper(self, fake_path1, fake_path2, *args, **kwargs):
        real_path1 = self.encode_real_path(self.get_real_path(
          self.decode_fake_path(fake_path1)))
        real_path2 = self.encode_real_path(self.get_real_path(
          self.decode_fake_path(fake_path2)))
        return real_op(real_path1, real_path2, *args, **kwargs)
    return wrapper

# For some reason, FuseOptParse overrides exit this such that we don't actually
# exit when optparse wants to.  This reverts to the original optparse behavior:
del FuseOptParse.exit

class FileSystemMappingToRealFilesOptionParser(FuseOptParse):
    def __init__(self, *args, **kwargs):
        kwargs['usage'] = '%prog [OPTIONS] {source} {mountpoint}'
        kwargs['version'] = '%%prog version %s' % version
        FuseOptParse.__init__(self, *args, **kwargs)

        self.add_option(
          mountopt = 'iocharset',
          default = DEFAULT_IOCHARSET,
          metavar = 'ENCODING',
          help = (
            'set mounted tree character encoding to ENCODING '
            '(default: %default)'
          ),
        )
        self.add_option(
          mountopt = 'source_iocharset',
          default = DEFAULT_SOURCE_IOCHARSET,
          metavar = 'ENCODING',
          help = (
            'set source directory character encoding to ENCODING '
            '(default: %default)'
          ),
        )
        self.add_option(
          mountopt = 'srcfilter',
          action = 'callback',
          type = str,
          callback = append_filter,
          metavar = 'EXPR',
          help = 'add a source path filter as specified by EXPR',
        )
        self.add_option(
          mountopt = 'dstfilter',
          action = 'callback',
          type = str,
          callback = append_filter,
          metavar = 'EXPR',
          help = 'add a destination path filter as specified by EXPR',
        )
        self.add_option(
          mountopt = 'verbosity',
          default = None,
          help = (
            'log file verbosity; one of debug, info, warning, error, critical '
            '(default: %s)'
          ) % DEFAULT_VERBOSITY,
        )
        self.add_option(
          mountopt = 'logsize',
          type = 'int',
          default = (1024 * 1024),
          help = 'length of log file ring buffer, in bytes',
        )

    def parse_args(self, *args, **kwargs):
        retval = FuseOptParse.parse_args(self, *args, **kwargs)
        if not hasattr(self.values, 'filters'):
            self.values.filters = [ ]
        return retval

fuse.feature_assert('stateful_files')

class File(object):
    _lock = None

    direct_io = True
    keep_cache = False

    # Undefined File methods:
    #
    # * lock

    def __init__(
      self, filesystem_instance, fake_path, flags, mode = None,
      truncate_to = None
    ):
        self._lock = RLock()
        self.filesystem_instance = filesystem_instance
        self.fake_path = fake_path
        self.real_path = filesystem_instance.get_real_path(fake_path)
        self.flags = flags
        self.mode = mode
        self.truncate_to = truncate_to

    def fgetattr(self): pass
    def read(self, length, offset): pass
    def release(self, flags): pass
    def write(self, buf, offset): pass
    def fsync(self, isfsyncfile): pass
    def flush(self): pass
    def ftruncate(self, len): pass

class ReadOnlyFile(File):
    direct_io = False

    fd = None
    file = None

    @exceptionally_logged(ignore = (PathNotFound,))
    @verbosely_logged
    def __init__(self, *args, **kwargs):
        super(ReadOnlyFile, self).__init__(*args, **kwargs)
        self.open_file()

    @nonatomic(attr_named('_lock'))
    def open_file(self):
        real_path = self.filesystem_instance.encode_real_path(self.real_path)
        args = [real_path, self.flags]
        if self.mode is not None:
            args.append(self.mode)
        self.file = os.fdopen(os.open(*args), 'r')

    @exceptionally_logged
    @logged
    @nonatomic(attr_named('_lock'))
    def read(self, length, offset):
        self.file.seek(offset)
        return self.file.read(length)

    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('_lock'))
    def fgetattr(self):
        return stat_result_to_fuse_stat(os.fstat(self.file.fileno()))

    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('_lock'))
    def release(self, flags):
        self.file.close()

    @exceptionally_logged
    @verbosely_logged
    def write(self, buf, offset):
        raise InvalidArgument

    @exceptionally_logged
    @verbosely_logged
    def ftruncate(self, len):
        raise InvalidArgument

class ReadWriteFile(File):
    fd = None

    @exceptionally_logged(ignore = (PathNotFound,))
    @verbosely_logged
    def __init__(self, *args, **kwargs):
        super(ReadWriteFile, self).__init__(*args, **kwargs)
        self.copy_path = os.path.join(
          os.path.dirname(self.real_path),
          u'.pytagsfs.%s' % os.path.basename(self.real_path)
        )
        self.open_file()

    @nonatomic(attr_named('_lock'))
    def open_file(self):
        flags = self.flags
        flags = flags | os.O_CREAT | os.O_EXCL

        real_path = self.filesystem_instance.encode_real_path(self.real_path)
        copy_path = self.filesystem_instance.encode_real_path(self.copy_path)

        args = [copy_path, flags]
        if self.mode is not None:
            args.append(self.mode)

        log_debug(
          u'ReadWriteFile.open_file: calling os.open with args %s' % repr(args))
        self.fd = os.open(*args)

        f = open(real_path, 'r')
        if self.truncate_to is None:
            data = f.read()
        else:
            data = f.read(self.truncate_to)
        f.close()

        os.write(self.fd, data)
        os.lseek(self.fd, 0, 0)

        shutil.copystat(real_path, copy_path)

    @exceptionally_logged
    @logged
    @nonatomic(attr_named('_lock'))
    def read(self, length, offset):
        os.lseek(self.fd, offset, 0)
        return os.read(self.fd, length)

    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('_lock'))
    def write(self, buf, offset):
        os.lseek(self.fd, offset, 0)
        return os.write(self.fd, buf)

    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('_lock'))
    def release(self, flags):
        retval = os.close(self.fd)
        real_path = self.filesystem_instance.encode_real_path(self.real_path)
        copy_path = self.filesystem_instance.encode_real_path(self.copy_path)
        os.rename(copy_path, real_path)
        return retval

    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('_lock'))
    def fsync(self, isfsyncfile):
        if isfsyncfile and hasattr(os, 'fdatasync'):
            return os.fdatasync(self.fd)
        else:
            return os.fsync(self.fd)

    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('_lock'))
    def flush(self):
        return os.close(os.dup(self.fd))

    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('_lock'))
    def fgetattr(self):
        return os.fstat(self.fd)

    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('_lock'))
    def ftruncate(self, len):
        return os.ftruncate(self.fd, len)

class ModeSelectFile(File):

    @exceptionally_logged(ignore = (PathNotFound,))
    @verbosely_logged
    def __init__(self, filesystem_instance, fake_path, flags, mode = None,
      truncate_to = None):
        if flags & (os.O_WRONLY | os.O_RDWR):
            cls = ReadWriteFile
        else:
            cls = ReadOnlyFile

        self.__class__ = cls
        cls.__init__(self, filesystem_instance, fake_path, flags, mode = mode,
          truncate_to = truncate_to)

class FileSystemMappingToRealFiles(Fuse):
    _lock = None

    source_tree_rep = None
    source_tree_path = None
    iocharset = None
    source_iocharset = None
    path_filters = None
    debug = False
    foreground = False
    readonly = None
    verbosity = None
    logsize = None

    truncated_paths = None
    write_locks = None
    write_locks_lock = None

    file_instances = None

    file_class = ModeSelectFile

    verbosities = {
      'debug': DEBUG,
      'info': INFO,
      'warning': WARNING,
      'error': ERROR,
      'critical': CRITICAL,
    }

    @exceptionally_logged
    @logged
    def __init__(self, *args, **kwargs):
        self._lock = RLock()

        kwargs['dash_s_do'] = 'setsingle'
        if 'parser_class' not in kwargs:
            kwargs['parser_class'] = FileSystemMappingToRealFilesOptionParser

        self.truncated_paths = {}
        self.write_locks = {}

        self.file_instances = {}

        VirtualLogFile.set_filesystem_instance(self)

        super(FileSystemMappingToRealFiles, self).__init__(*args, **kwargs)

    fuse.feature_assert('has_init')
    @exceptionally_logged
    @logged
    def fsinit(self):
        log_warning(u'pytagsfs version %s' % version)

        # This is called from the end of pre_init so that the source tree
        # representation gets populated before the process detaches from the
        # user's shell.  Otherwise, the initial population delay is observed
        # when the filesystem is first accessed, and this is confusing to users,
        # who have no way to tell if the filesystem has simply hung.
        #self.source_tree_rep.start()

    # This clean-up is performed at the end of main to work around what appears
    # to be a bug in python-fuse that prevents exit due to a GIL deadlock.  The
    # bug is fixed upstream, but the fix may not have been released yet.

    #fuse.feature_assert('has_destroy')
    #@exceptionally_logged
    #def fsdestroy(self):
    #    self.source_tree_rep.stop()

    @exceptionally_logged
    @logged
    def main(self):
        try:
            self.parse()
        except OptParseError, e:
            self.parser.error(unicode(e))

        try:
            retval = super(FileSystemMappingToRealFiles, self).main()
        except fuse.FuseError, e:
            retval = 1
            log_error(unicode(e))
        else:
            self.source_tree_rep.stop()
        return retval

    def parse(self, *args, **kwargs):
        retval = super(FileSystemMappingToRealFiles, self).parse(
          *args, **kwargs)
        self.process_options()
        self.pre_init()
        return retval

    def process_options(self):
        opts, args = self.cmdline

        if 'debug' in self.parser.fuse_args.optlist:
            self.debug = True
            self.foreground = True
            self.verbosity = self.verbosities['debug']

        if self.parser.fuse_args.modifiers['foreground']:
            self.foreground = True

        if 'ro' in self.parser.fuse_args.optlist:
            self.readonly = True

        if self.parser.values.verbosity is not None:
            if self.debug:
                log_warning(
                  u'warning: option debug overrides user-specified verbosity')
            else:
                try:
                    self.verbosity = self.verbosities[
                      self.parser.values.verbosity]
                except KeyError:
                    raise OptParseError(
                      'verbosity should be one of %s' % ', '.join(
                        self.verbosities.keys())
                    )
        elif not self.debug:
            self.verbosity = self.verbosities[DEFAULT_VERBOSITY]

        self.logsize = self.parser.values.logsize

        if len(args) < 1:
            raise OptParseError('too few arguments.')

        encoding = locale.getpreferredencoding()

        self.source_tree_path = os.path.abspath(unicode(args[0], encoding))
        self.iocharset = self.parser.values.iocharset
        self.source_iocharset = self.parser.values.source_iocharset

        self.path_filters = [ ]
        for which, expr in self.parser.values.filters:
            real = (which == 'src')
            self.path_filters.append((expr, real))

    def pre_init(self):
        set_logsize(self.logsize)

        set_log_level(self.verbosity)
        if self.foreground:
            enable_stderr()

        self.source_tree_rep = self.build_source_tree_rep()
        self.multithreaded = (
          self.multithreaded and
          self.source_tree_rep.supports_threads()
        )
        log_info(u'multithreaded = %s' % self.multithreaded)

        self.source_tree_rep.start()

    def build_source_tree_rep(self):
        raise NotImplementedError

################################################################################

    def put_file(self, real_path):
        self.assertUnicode(real_path)
        return self.source_tree_rep.put_file(real_path)

    def delete_file(self, fake_path):
        self.assertUnicode(fake_path)
        return self.source_tree_rep.delete_file(fake_path)

    def rename_path(self, old_fake_path, new_fake_path):
        self.assertUnicode(old_fake_path)
        self.assertUnicode(new_fake_path)
        return self.source_tree_rep.rename_path(old_fake_path, new_fake_path)

    def create_directory(self, fake_path):
        self.assertUnicode(fake_path)
        return self.source_tree_rep.create_directory(fake_path)

    def remove_directory(self, fake_path):
        self.assertUnicode(fake_path)
        return self.source_tree_rep.remove_directory(fake_path)

    def get_real_path(self, fake_path):
        self.assertUnicode(fake_path)
        return self.source_tree_rep.get_real_path(fake_path)

    def get_fake_path(self, real_path):
        self.assertUnicode(real_path)
        return self.source_tree_rep.get_fake_path(real_path)

    def get_entries(self, fake_path):
        self.assertUnicode(fake_path)
        return self.source_tree_rep.get_entries(fake_path)

    def path_exists(self, fake_path):
        self.assertUnicode(fake_path)
        return self.source_tree_rep.path_exists(fake_path)

    def is_file(self, fake_path):
        self.assertUnicode(fake_path)
        return self.source_tree_rep.is_file(fake_path)

    def is_dir(self, fake_path):
        self.assertUnicode(fake_path)
        return self.source_tree_rep.is_dir(fake_path)

    @nonatomic(attr_named('_lock'))
    def write_lock_path(self, fake_path):
        self.assertUnicode(fake_path)
        lock = self.write_locks.get(fake_path, Lock())
        self.write_locks[fake_path] = lock
        lock.acquire()

    @nonatomic(attr_named('_lock'))
    def write_unlock_path(self, fake_path):
        self.assertUnicode(fake_path)
        self.write_locks[fake_path].release()

    def assertUnicode(self, s):
        if not isinstance(s, unicode):
            raise AssertionError('object is not a unicode string: %s' % repr(s))

    def encode_fake_path(self, s):
        return s.encode(self.iocharset)

    def decode_fake_path(self, s):
        return s.decode(self.iocharset)

    def encode_real_path(self, s):
        return self.source_tree_rep.source_tree.encode(s)

    def decode_real_path(self, s):
        return self.source_tree_rep.source_tree.decode(s)

################################################################################

    @exceptionally_logged
    @verbosely_logged
    @return_errno
    def getattr(self, fake_path):
        fake_path = self.decode_fake_path(fake_path)
        try:
            return self.file_instances[fake_path].fgetattr()
        except KeyError:
            return stat_result_to_fuse_stat(
              self.source_tree_rep.getattr(fake_path))

    chmod = exceptionally_logged(operation_on_one_real_path(os.chmod))
    chown = exceptionally_logged(operation_on_one_real_path(os.chown))

    @exceptionally_logged
    @verbosely_logged
    @return_errno
    def truncate(self, fake_path, len):
        fake_path = self.decode_fake_path(fake_path)
        log_debug(u'truncate: write-locking path %s' % fake_path)
        self.write_lock_path(fake_path)
        try:
            try:
                if len < self.truncated_paths[fake_path]:
                    self.truncated_paths[fake_path] = len
            except KeyError:
                self.truncated_paths[fake_path] = len
        finally:
            log_debug(u'truncate: write-unlocking path %s' % fake_path)
            self.write_unlock_path(fake_path)

    @exceptionally_logged
    @verbosely_logged
    @return_errno
    @nonatomic(attr_named('_lock'))
    def open(self, fake_path, flags):
        fake_path = self.decode_fake_path(fake_path)
        if (os.O_RDWR | os.O_WRONLY) & flags:
            log_debug(u'open: write-locking path %s' % fake_path)
            self.write_lock_path(fake_path)
        file_instance = ModeSelectFile(
          self, fake_path, flags,
          truncate_to = self.truncated_paths.get(fake_path, None)
        )
        if (os.O_RDWR | os.O_WRONLY) & flags:
            self.file_instances[fake_path] = file_instance
        return file_instance

    @exceptionally_logged
    @verbosely_logged
    @return_errno
    @nonatomic(attr_named('_lock'))
    def release(self, fake_path, flags, file_obj):
        fake_path = self.decode_fake_path(fake_path)
        try:
            del self.truncated_paths[fake_path]
        except KeyError:
            pass
        retval = Fuse.__getattr__(self, 'release')(fake_path, flags, file_obj)
        if (os.O_RDWR | os.O_WRONLY) & flags:
            log_debug(u'release: write-unlocking path %s' % fake_path)
            self.write_unlock_path(fake_path)
        try:
            del self.file_instances[fake_path]
        except KeyError:
            pass
        return retval

    @exceptionally_logged
    @verbosely_logged
    @return_errno
    def create(self, fake_path, flags, mode):
        raise InvalidArgument

    @exceptionally_logged
    @verbosely_logged
    @return_errno
    def utime(self, fake_path, times):
        fake_path = self.decode_fake_path(fake_path)
        self.source_tree_rep.utime(fake_path, times)
        return 0

    @exceptionally_logged
    @verbosely_logged
    @return_errno
    def statfs(self):
        return os.statvfs(self.source_tree_path)

    @exceptionally_logged
    @logged
    def readdir(self, fake_path, offset):
        fake_path = self.decode_fake_path(fake_path)
        try:
            entries = [
              self.encode_fake_path(e) for e in self.get_entries(fake_path)]
        except PathNotFound:
            return -errno.ENOENT
        return [fuse.Direntry(e) for e in entries]

    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('_lock'))
    def access(self, fake_path, mode):
        fake_path = self.decode_fake_path(fake_path)
        try:
            real_path = self.get_real_path(fake_path)
        except PathNotFound:
            try:
                self.get_entries(fake_path)
            except PathNotFound:
                return -errno.ENOENT
        else:
            if not os.access(self.encode_real_path(real_path), mode):
                return -errno.EACCES
        return 0

    @exceptionally_logged
    @verbosely_logged
    @return_errno
    @nonatomic(attr_named('_lock'))
    def rename(self, old_fake_path, new_fake_path):
        old_fake_path = self.decode_fake_path(old_fake_path)
        new_fake_path = self.decode_fake_path(new_fake_path)
        self.rename_path(old_fake_path, new_fake_path)
        return 0

    @exceptionally_logged
    @verbosely_logged
    @return_errno
    @nonatomic(attr_named('_lock'))
    def mkdir(self, fake_path, mode):
        fake_path = self.decode_fake_path(fake_path)
        self.create_directory(fake_path)
        return 0

    @exceptionally_logged
    @verbosely_logged
    @return_errno
    @nonatomic(attr_named('_lock'))
    def rmdir(self, fake_path):
        fake_path = self.decode_fake_path(fake_path)
        self.remove_directory(fake_path)
        return 0

    # Undefined FUSE methods:
    #
    # * setxattr
    # * getxattr
    # * listxattr
    # * removexattr
    # * opendir
    # * releasedir
    # * fsyncdir
    # * init
    # * destroy
    # * utimens
    # * bmap
    # * readlink
    # * mknod
    # * unlink
    # * symlink
    # * link

################################################################################

DEFAULT_FORMAT = '/%f'
DEFAULT_PATH_STORE = \
  'pytagsfs.pathstore.pytypes.PyTypesPathStore'
DEFAULT_SOURCE_TREE_REPRESENTATION = \
  'pytagsfs.sourcetreerep.pollinline.PollInLineSourceTreeRepresentation'

SOURCE_TREE_MONITORS = (
  'pytagsfs.sourcetreemon.pyinotify_.DeferredPyInotifySourceTreeMonitor',
  'pytagsfs.sourcetreemon.gamin_.DeferredGaminSourceTreeMonitor',
  'pytagsfs.sourcetreemon.dummy.DummySourceTreeMonitor',
)

class PyTagsFileSystemOptionParser(FileSystemMappingToRealFilesOptionParser):
    def __init__(self, *args, **kwargs):
        FileSystemMappingToRealFilesOptionParser.__init__(self, *args, **kwargs)
        self.add_option(
          mountopt = 'sourcetreerep',
          default = DEFAULT_SOURCE_TREE_REPRESENTATION,
          help = SUPPRESS_HELP,
        )
        self.add_option(
          mountopt = 'pathstore',
          default = DEFAULT_PATH_STORE,
          help = SUPPRESS_HELP,
        )
        self.add_option(
          mountopt = 'sourcetreemon',
          default = None,
          help = SUPPRESS_HELP,
        )
        self.add_option(
          mountopt = 'format',
          default = DEFAULT_FORMAT,
          metavar = 'FORMAT',
          help = (
            'set destination path format string '
            '(default: %default)'
          ),
        )

class PyTagsFileSystem(FileSystemMappingToRealFiles):
    source_tree_rep_cls_dotted_name = None
    path_store_cls_dotted_name = None
    source_tree_mon_cls_dotted_name = None
    format_string = None

    def __init__(self, *args, **kwargs):
        if 'parser_class' not in kwargs:
            kwargs['parser_class'] = PyTagsFileSystemOptionParser
        super(PyTagsFileSystem, self).__init__(*args, **kwargs)

    def process_options(self, *args, **kwargs):
        super(PyTagsFileSystem, self).process_options(*args, **kwargs)

        encoding = locale.getpreferredencoding()

        self.source_tree_rep_cls_dotted_name = self.parser.values.sourcetreerep
        self.path_store_cls_dotted_name = self.parser.values.pathstore
        self.source_tree_mon_cls_dotted_name = self.parser.values.sourcetreemon
        self.format_string = unicode(self.parser.values.format, encoding)

        self.format_string = self.format_string.rstrip('/')
        if not self.format_string.startswith('/'):
            self.format_string = '/%s' % self.format_string
        self.format_string = re.sub(r'/{2,}', '/', self.format_string)

    def build_source_tree_rep(self):
        source_tree_rep_cls = get_obj_by_dotted_name(
          self.source_tree_rep_cls_dotted_name)
        path_store_cls = get_obj_by_dotted_name(
          self.path_store_cls_dotted_name)

        log_info(u'source_tree_rep_cls = %s' % repr(source_tree_rep_cls))
        log_info(u'path_store_cls = %s' % repr(path_store_cls))

        if self.source_tree_mon_cls_dotted_name is not None:
            source_tree_mon_cls_dotted_names = [
              self.source_tree_mon_cls_dotted_name]
        else:
            source_tree_mon_cls_dotted_names = list(SOURCE_TREE_MONITORS)

        monitor = None
        for source_tree_mon_cls_dotted_name in (
          source_tree_mon_cls_dotted_names):
            log_debug(
              'trying source tree monitor class %s' %
              source_tree_mon_cls_dotted_name
            )
            source_tree_mon_cls = get_obj_by_dotted_name(
              source_tree_mon_cls_dotted_name)
            log_debug('source_tree_mon_cls = %s' % repr(source_tree_mon_cls))
            try:
                monitor = source_tree_mon_cls()
                log_debug('monitor = %s' % repr(monitor))
            except MissingDependency, e:
                log_warning(
                  (
                    u'WARNING: not using source tree monitor %s due to '
                    u'missing dependency %s'
                  ) % (
                    source_tree_mon_cls.__name__,
                    unicode(e),
                  )
                )
            else:
                break
        if monitor is None:
            raise OptParseError('No supported source tree monitor found.')

        log_info(u'source_tree_mon_cls = %s' % repr(source_tree_mon_cls))

        if (not self.readonly) and (not monitor.supports_writes()):
            raise OptParseError(
              'source tree monitor does not support writes, use -o ro.')

        meta_parser = DelegateMultiMetaParser((
          FilenameMetaParser(),
          ParentDirMetaParser(),
          ExtensionMetaParser(),
          MutagenFileMetaParser(),
        ))
        return source_tree_rep_cls(
          path_parser = PathParser(meta_parser, self.format_string),
          path_store = path_store_cls(),
          source_tree = SourceTree(
            root = self.source_tree_path,
            iocharset = self.source_iocharset,
          ),
          monitor = monitor,
          cache = PathPropCache(),
          filters = self.path_filters,
          debug = self.debug,
        )
