# Copyright (c) 2007-2008 Forest Bond.
# This file is part of the pytagsfs software package.
#
# pytagsfs is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import os.path

from pytagsfs.debug import log_error
from pytagsfs.exceptions import InvalidArgument

class MetaParser(object):
    def get_fields(self):
        raise NotImplementedError

    def parse(self, path):
        raise NotImplementedError

    def apply(self, path, d, dry_run = False):
        '''
        Change the file at ``path`` such that parsing meta-data from the file
        would result in a dictionary with values matching those of ``d``.
        
        The default implementation raises InvalidArgument if changes to the
        file are actually necessary.  Sub-class implementations should override
        this and actually make those changes where appropriate.

        Implementations should *never* make the file unchangeable by subsequent
        calls to apply on other sub-classes.  For instance, renaming the file
        would most certainly cause this.
        '''
        d_parse = self.parse(path)
        for k in d_parse:
            try:
                if d_parse[k] != d[k]:
                    raise InvalidArgument
            except KeyError:
                pass

class FilenameMetaParser(MetaParser):
    '''
    >>> p = FilenameMetaParser()

    >>> p.get_fields()
    ['f']

    >>> p.parse(os.path.join('a', 'b', 'c'))
    {'f': 'c'}

    >>> p.parse('/')
    {}

    >>> p.parse('')
    {}
    '''

    def get_fields(self):
        return ['f']

    def parse(self, path):
        d = { }
        f = os.path.basename(path.rstrip('/'))
        if f:
            d['f'] = f
        return d

class ParentDirMetaParser(MetaParser):
    '''
    >>> p = ParentDirMetaParser()

    >>> p.get_fields()
    ['p']

    >>> p.parse(os.path.join('a', 'b', 'c'))
    {'p': 'b'}

    >>> p.parse('/')
    {}

    >>> p.parse('')
    {}
    '''

    def get_fields(self):
        return ['p']

    def parse(self, path):
        d = { }
        p = os.path.basename(os.path.dirname(path.rstrip('/')))
        if p:
            d['p'] = p
        return d

class ExtensionMetaParser(MetaParser):
    '''
    >>> p = ExtensionMetaParser()

    >>> p.get_fields()
    ['e']

    >>> p.parse(os.path.join('a', 'b', 'foo.bar'))
    {'e': 'bar'}

    >>> p.parse('/')
    {}

    >>> p.parse('foo')
    {}

    >>> p.parse('')
    {}
    '''

    def get_fields(self):
        return ['e']

    def parse(self, path):
        d = { }
        parts = os.path.basename(path.rstrip('/')).rsplit('.', 1)
        if len(parts) > 1:
            d['e'] = parts[-1]
        return d

def _merge_lists_uniq(*lists):
    '''
    >>> l1 = [1, 2, 3, 3]
    >>> l2 = [2, 4, 5, 7]
    >>> _merge_lists_uniq(l1, l2)
    [1, 2, 3, 4, 5, 7]
    '''
    result = [ ]
    for l in lists:
        result.extend(l)
    return list(set(result))

def _merge_dicts(*dicts):
    '''
    >>> d1 = {'a': 'x', 'b': 'y'}
    >>> d2 = {'b': 'm', 'c': 'n'}
    >>> d3 = _merge_dicts(d1, d2)
    >>> [(k, d3[k]) for k in sorted(d3.keys())]
    [('a', 'x'), ('b', 'm'), ('c', 'n')]
    '''
    result = { }
    for d in dicts:
        result.update(d)
    return result

class DelegateMultiMetaParser(MetaParser):
    '''
    >>> p = DelegateMultiMetaParser([
    ...   FilenameMetaParser(),
    ...   ParentDirMetaParser(),
    ... ])

    >>> sorted(p.get_fields())
    ['f', 'p']

    >>> d = p.parse(os.path.join('a', 'b', 'c'))
    >>> [(k, d[k]) for k in sorted(d.keys())]
    [('f', 'c'), ('p', 'b')]

    >>> d = p.parse(os.path.join('a'))
    >>> [(k, d[k]) for k in sorted(d.keys())]
    [('f', 'a')]

    >>> d = p.parse('/')
    >>> [(k, d[k]) for k in sorted(d.keys())]
    []

    >>> d = p.parse('')
    >>> [(k, d[k]) for k in sorted(d.keys())]
    []
    '''

    def __init__(self, parsers):
        self.parsers = parsers

    def get_fields(self):
        field_lists = [ p.get_fields() for p in self.parsers ]
        return _merge_lists_uniq(*field_lists)

    def parse(self, path):
        meta_data_dicts = [ p.parse(path) for p in self.parsers ]
        return _merge_dicts(*meta_data_dicts)

    def apply(self, path, d, dry_run = False):
        for _dry_run in (True, dry_run):
            for parser in reversed(self.parsers):
                try:
                    parser.apply(path, d, dry_run = _dry_run)
                except NotImplementedError:
                    pass

MUTAGEN_FIELD_MAPPING = {
  'n': 'tracknumber',
  'a': 'artist',
  't': 'title',
  'l': 'album',
  'y': 'date',
  'g': 'genre',
}

def get_field_for_mutagen_field(mutagen_field):
    for field in MUTAGEN_FIELD_MAPPING:
        if MUTAGEN_FIELD_MAPPING[field] == mutagen_field:
            return field
    raise ValueError('No such mutagen field: %s' % mutagen_field)

class _BaseMutagenMetaParser(MetaParser):
    tags_class = None
    error_class = None

    def __init__(self):
        self._mutagen_fields = MUTAGEN_FIELD_MAPPING.keys()
        self._other_fields = ['N']

    def get_fields(self):
        return self._mutagen_fields + self._other_fields

    def make_tags_obj(self, path):
        cls = self.tags_class
        return cls(path)

    def parse(self, path):
        try:
            tags = self.make_tags_obj(path)
        except self.error_class:
            tags = None
        if tags is None:
            tags = { }

        d = { }

        for field in self._mutagen_fields:
            mutagen_field = MUTAGEN_FIELD_MAPPING[field]
            if tags.get(mutagen_field, False):
                d[field] = tags[mutagen_field][0]

        self._post_process(d)

        return d

    def _post_process(self, d):
        if 'n' in d:
            try:
                d['n'] = '%u' % int(d['n'])
            except ValueError:
                del d['n']
            else:
                d['N'] = '%02u' % int(d['n'])

    def apply(self, path, d, dry_run = False):
        tags = self.make_tags_obj(path)

        for field in self._mutagen_fields:
            mutagen_field = MUTAGEN_FIELD_MAPPING[field]
            if field in d:
                tags[mutagen_field] = d[field]

        if not dry_run:
            tags.save()

from pytagsfs.util import SimpleMutagenFile

class MutagenFileMetaParser(_BaseMutagenMetaParser):
    tags_class = staticmethod(SimpleMutagenFile)
    error_class = Exception

from mutagen.easyid3 import EasyID3, error as easyid3_error

class MutagenEasyID3MetaParser(_BaseMutagenMetaParser):
    tags_class = EasyID3
    error_class = easyid3_error

from mutagen.flac import error as flac_error, FLAC

class MutagenFLACMetaParser(_BaseMutagenMetaParser):
    tags_class = FLAC
    error_class = flac_error
