# Copyright (c) 2007-2008 Forest Bond.
# This file is part of the pytagsfs software package.
#
# pytagsfs is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

class PathStore(object):
    '''
    Abstract base class for path store implementations.

    Sub-classes must override the following methods:

     * ``put_file``
     * ``delete_file``
     * ``rename_file``
     * ``get_real_path``
     * ``get_fake_path``
     * ``get_entries``
     * ``is_file``
     * ``is_dir``
     * ``path_exists``

    Path stores store path mappings from real paths to fake (virtual) paths.
    Multiple real paths can be associated with a given fake path.  The most
    recently added real path takes precedence over previously added real paths.
    Conceptually, associating a real path with a fake path should behave like a
    stack, such that:

    >>> path_store.put_file('/foo', '/bar')
    >>> path_store.get_real_path('/foo')
    '/bar'
    >>> path_store.put_file('/foo', '/baz')
    >>> path_store.get_real_path('/foo')
    '/baz'
    >>> path_store.delete_file('/foo')
    >>> path_store.get_real_path('/foo')
    '/bar'
    >>> path_store.delete_file('/foo')
    >>> path_store.get_real_path('/foo')
    Traceback (most recent call last):
    ...
    FakePathNotFound: /foo
    
    Parent directories of virtual files are implicitly created.  Once all files
    and sub-directories of these directories cease to exist, the directories
    themselves should magically disappear:

    >>> path_store.put_file('/foo/bar', '/baz')
    >>> path_store.get_entries('/foo')
    ['bar']
    >>> path_store.delete_file('/foo/bar')
    >>> path_store.get_entries('/foo')
    Traceback (most recent call last):
    ...
    FakePathNotFound: /foo

    If the path store implementation supports tracking of empty directories,
    the following methods should be overridden:

     * ``create_directory``
     * ``remove_directory``

    Empty directories are handled differently from other directories in that
    they must be explicitly created, and that they are transient, which is to
    say that they cease to exist if they at any point contain entries and those
    entries are removed via ``delete_file`` or ``remove_directory``.

    >>> path_store.create_directory('/foo')
    >>> path_store.get_entries('/foo')
    []
    >>> path_store.put_file('/foo/bar/baz', '/boink')
    >>> path_store.get_entries('/foo')
    ['bar']
    >>> path_store.delete_file('/foo/bar/baz')
    >>> path_store.get_entries('/foo')
    Traceback (most recent call last):
    ...
    FakePathNotFound: /foo

    >>> path_store.create_directory('/foo')
    >>> path_store.get_entries('/foo')
    []
    >>> path_store.create_directory('/foo/bar')
    >>> path_store.get_entries('/foo')
    ['bar']
    >>> path_store.get_entries('/foo/bar')
    []
    >>> path_store.remove_directory('/foo/bar')
    >>> path_store.get_entries('/foo')
    Traceback (most recent call last):
    ...
    FakePathNotFound: /foo

    Another way of thinking about this is that empty directories exist as end
    points only in order to support incrementally building longer paths.  Once
    the need for an empty end point subsides, these directories are handled
    exactly the same as other implicitly-created directories.

    Note that it is perfectly acceptable for implementations to not override
    ``create_directory`` and ``remove_directory``.  The default implementations
    raise ``NotImplementedError``, which should be handled correctly by
    callers.  However, if one of either of these two methods is implemented,
    both should be in order to provide users with symmetric functionality.

    If the path store is thread safe, it should also override method
    ``supports_threads`` and return True (possibly only conditionally for the
    duration of the mount).

    All paths (both fake and real) are specified either as ASCII byte-strings
    or as Unicode strings.  Non-ASCII byte-strings need not be handled.

    Sub-classes are encouraged to define their class-level docstrings as
    follows in order to test consistency with the assumptions expressed here:

    # class MyPathStoreImplementation(PathStore):
    #     \'\'\'
    #     Documentation for MyPathStoreImplementation.
    #
    #     >>> path_store = MyPathStoreImplementation()
    #
    #     \'\'\' + PathStore.__doc__

    Note that implementations that don't override ``create_directory`` and
    ``remove_directory`` will fail some of these examples with
    ``NotImplementedError``.
    '''

    def put_file(self, fake_path, real_path):
        '''
        Add path mapping ``fake_path`` => ``real_path`` to the path store.
        Note that multiple real paths can be associated with any given fake
        path, with the most recently-added real path being returned by
        ``get_real_path``.  See the class docstring for more information.
        '''
        raise NotImplementedError

    def delete_file(self, fake_path):
        '''
        Remove path mapping by fake path.  See the class docstring for more
        information.
        '''
        raise NotImplementedError

    def rename_file(self, old_fake_path, new_fake_path):
        '''
        Rename stored fake path ``old_fake_path`` to ``new_fake_path``.  This
        is subtly different from calling ``delete_file(old_fake_path)``
        followed by ``put_file(new_fake_path, real_path)`` in that
        ``rename_file`` should not cause the order of entries returned by
        ``get_entries`` to change when ``old_fake_path`` and ``new_fake_path``
        refer to virtual files within the same virtual directory.
        additionally, ``rename_file`` should always be more efficient than this
        remove/add pattern.
        '''
        raise NotImplementedError

    def create_directory(self, fake_path):
        '''
        Create an empty virtual directory by virtual path.  Parent directories
        are implicitly created as for ``put_file``.  If the virtual path already
        exists in the path store, ``PathExists`` should be raised.
        '''
        raise NotImplementedError

    def remove_directory(self, fake_path):
        '''
        Remove an empty directory by virtual path.  If the such a directory
        exists but it is not an empty directory, ``DirectoryNotEmpty`` should
        be raised.  Raise ``FakePathNotFound`` if the fake path does not exist
        in the path store.
        '''
        raise NotImplementedError

    def get_real_path(self, fake_path):
        '''
        Return the most recently added real path associated with virtual path
        ``fake_path``.  Raise ``FakePathNotFound`` if this virtual path is not
        in the path store.
        '''
        raise NotImplementedError

    def get_fake_path(self, real_path):
        '''
        Return the fake path associated with real path ``real_path``.  Raise
        ``RealPathNotFound`` if no such association exists in the path store.
        '''
        raise NotImplementedError

    def get_entries(self, fake_path):
        '''
        Return the names of all file and directories contained by the virtual
        directory with virtual path ``fake_path``.  The order of these entries
        should be consistent across multiple calls to ``get_entries``.  Raise
        ``NotADirectory`` if ``fake_path`` refers to a virtual file rather than
        a virtual directory.  Raise ``FakePathNotFound`` if no such fake path
        exists in the path store.
        '''
        raise NotImplementedError

    def is_file(self, fake_path):
        '''
        Return True if virtual path ``fake_path`` refers to a virtual file.
        Return False otherwise, even if ``fake_path`` does not exist in the
        path store.
        '''
        raise NotImplementedError

    def is_dir(self, fake_path):
        '''
        Return True if virtual path ``fake_path`` refers to a virtual
        directory, whether that directory is empty or not.  Return False
        otherwise, even if ``fake_path`` does not exist in the path store.
        '''
        raise NotImplementedError

    def path_exists(self, fake_path):
        '''
        Return True if ``fake_path`` exists in the path store, independent of
        whether that path refers to a file or directory (empty or not).  Return
        False otherwise, even if ``fake_path`` does not exist in the path
        store.
        '''
        raise NotImplementedError

    def supports_threads(self):
        '''
        Return True if the path store is thread-safe under current run-time
        conditions.  The result of this call should not vary over the course of
        a single mount.  In most cases, it will unconditionally return True or
        False to indicate the thread-safety of the specific path store
        implementation.  The default implementation returns False
        unconditionally.
        '''
        return False
