# Copyright (c) 2007-2008 Forest Bond.
# This file is part of the pytagsfs software package.
#
# pytagsfs is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import os.path

try:
    from functools import wraps
except ImportError:
    from sclapp.legacy_support import wraps

from sclapp.util import safe_encode

from threading import RLock

from pytagsfs.pathstore import PathStore
from pytagsfs.util import last_unique, nonatomic, attr_named
from pytagsfs.exceptions import (
  FakePathNotFound,
  RealPathNotFound,
  NotADirectory,
  PathExists,
  PathNotFound,
  InvalidArgument,
  DirectoryNotEmpty,
)

class TwoWayMapping(object):
    '''
    >>> m = TwoWayMapping()
    >>> m.add('foo', 'bar')
    >>> m.add('baz', 'bam')
    >>> m.lookup('foo')
    'bar'
    >>> m.lookup('baz')
    'bam'
    >>> m.reverse_lookup('bar')
    'foo'
    >>> m.reverse_lookup('bam')
    'baz'
    '''
    def __init__(self):
        self.forward = {}
        self.reverse = {}

    def add(self, fval, rval):
        self.forward[fval] = rval
        self.reverse[rval] = fval

    def remove(self, fval):
        del self.reverse[self.forward[fval]]
        del self.forward[fval]

    def lookup(self, fval):
        return self.forward[fval]

    def reverse_lookup(self, rval):
        return self.reverse[rval]

class DictList(dict):
    '''
    >>> dl = DictList()
    >>> dl['foo'].append(0)
    >>> dl['foo'].append(1)
    >>> dl['foo'].append(2)
    >>> dl['foo']
    [0, 1, 2]
    >>> dl['foo'].pop()
    2
    >>> dl['foo'].pop()
    1
    >>> dl['foo'].pop()
    0
    >>> dl['bar'] = 3
    Traceback (most recent call last):
    ...
    TypeError
    '''
    def __setitem__(self, k, v):
        raise TypeError

    def __getitem__(self, k):
        if k not in self:
            super(DictList, self).__setitem__(k, [])
        return dict.__getitem__(self, k)

class DictListWithExplicitKeys(dict):
    '''
    >>> dl = DictListWithExplicitKeys()
    >>> dl['foo']
    Traceback (most recent call last):
    ...
    KeyError: 'foo'
    >>> dl.define('foo')
    >>> dl['foo']
    []
    >>> dl.undefine('foo')
    >>> dl.undefine('foo')
    Traceback (most recent call last):
    ...
    KeyError: 'foo'
    >>> dl['foo'] = []
    Traceback (most recent call last):
    ...
    TypeError: __setitem__ unsupported; use define
    >>> dl.define('foo')
    >>> del dl['foo']
    Traceback (most recent call last):
    ...
    TypeError: __delitem__ unsupported; use undefine
    '''
    def __setitem__(self, k, v):
        raise TypeError('__setitem__ unsupported; use define')

    def __delitem__(self, k):
        raise TypeError('__delitem__ unsupported; use undefine')

    def define(self, k):
        if k in self:
            raise KeyError('key %s already exists' % k)
        dict.__setitem__(self, k, [])

    def undefine(self, k):
        dict.__delitem__(self, k)

class CallbackStack(object):
    '''
    >>> def print_cb(*args):
    ...     print repr(args)
    >>> stack = CallbackStack(print_cb, print_cb)

    >>> stack.push('foo')
    (0, 'foo')
    >>> stack.push('bar')
    (1, 'bar')
    >>> stack.push('baz')
    (2, 'baz')
    >>> stack.data
    ['foo', 'bar', 'baz']
    >>> stack.peek()
    'baz'
    >>> stack.pop()
    (2, 'baz')
    'baz'
    >>> stack.peek()
    'bar'
    >>> stack.pop()
    (1, 'bar')
    'bar'
    >>> stack.peek()
    'foo'
    >>> stack.pop()
    (0, 'foo')
    'foo'
    >>> stack.pop()
    Traceback (most recent call last):
    ...
    IndexError: pop from empty list
    '''
    def __init__(self, push_cb = None, pop_cb = None, data = ()):
        self.data = list(data)
        self.push_cb = push_cb
        self.pop_cb = pop_cb

    def __unicode__(self):
        return u'CallbackStack with data %s' % repr(self.data)

    def __str__(self):
        return safe_encode(unicode(self))

    def __repr__(self):
        return safe_encode(
          'CallbackStack(push_cb = %s, pop_cb = %s, data = %s)' % (
            repr(self.push_cb), repr(self.pop_cb), repr(self.data)
          )
        )

    def push(self, item):
        if self.push_cb is not None:
            self.push_cb(len(self.data), item)
        self.data.append(item)

    def pop(self):
        item = self.data.pop()
        if self.pop_cb is not None:
            self.pop_cb(len(self.data), item)
        return item

    def peek(self):
        return self.data[-1]

class DictStackWithReverseLookupsWithExplicitKeys(DictList):
    '''
    >>> d = DictStackWithReverseLookupsWithExplicitKeys()

    >>> d.define('foo')

    >>> d['foo'].__class__.__name__, d['foo'].data
    ('CallbackStack', [])
    >>> d.lookup('foo')
    Traceback (most recent call last):
    ...
    KeyError: 'foo'

    >>> d['foo'].push('foo_one')
    >>> d.lookup('foo')
    'foo_one'
    >>> d.reverse_lookup('foo_one')
    'foo'

    >>> d['foo'].push('foo_two')
    >>> d.lookup('foo')
    'foo_two'
    >>> d.reverse_lookup('foo_two')
    'foo'
    >>> d.reverse_lookup('foo_one')
    Traceback (most recent call last):
    ...
    KeyError: 'foo_one'

    >>> d['foo'].push('foo_three')
    >>> d.lookup('foo')
    'foo_three'
    >>> d.reverse_lookup('foo_three')
    'foo'
    >>> d.reverse_lookup('foo_two')
    Traceback (most recent call last):
    ...
    KeyError: 'foo_two'
    >>> d.reverse_lookup('foo_one')
    Traceback (most recent call last):
    ...
    KeyError: 'foo_one'

    >>> d['foo'].pop()
    'foo_three'
    >>> d['foo'].pop()
    'foo_two'
    >>> d['foo'].pop()
    'foo_one'

    >>> d.lookup('foo')
    Traceback (most recent call last):
    ...
    KeyError: 'foo'
    >>> d['foo'].__class__.__name__, d['foo'].data
    ('CallbackStack', [])

    >>> d.undefine('foo')
    >>> d['foo']
    Traceback (most recent call last):
    ...
    KeyError: 'foo'
    '''
    def __init__(self):
        super(DictStackWithReverseLookupsWithExplicitKeys, self).__init__()
        self.reverse_mapping = {}

    def __setitem__(self, k, v):
        raise TypeError('__setitem__ not supported')

    def __getitem__(self, k):
        return dict.__getitem__(self, k)

    def __delitem__(self, k):
        raise TypeError('__delitem__ not supported')

    def define(self, k):
        if k in self:
            raise KeyError('key %s already exists' % k)
        dict.__setitem__(
          self, k, CallbackStack(
            push_cb = self.make_push_cb(k),
            pop_cb = self.make_pop_cb(k),
          ),
        )

    def undefine(self, k):
        dict.__delitem__(self, k)

    def make_push_cb(self, k):
        def push_cb(index, item):
            if index > 0:
                old_item = self[k].data[-1]
                del self.reverse_mapping[old_item]
            self.reverse_mapping[item] = k
        return push_cb

    def make_pop_cb(self, k):
        def pop_cb(index, item):
            del self.reverse_mapping[item]
            if index > 0:
                old_item = self[k].data[-1]
                self.reverse_mapping[old_item] = k
        return pop_cb

    def lookup(self, k):
        # FIXME: is KeyError really appropriate here?  It's better than
        # IndexError, I guess.
        try:
            return self[k].peek()
        except IndexError:
            raise KeyError(k)

    def reverse_lookup(self, v):
        return self.reverse_mapping[v]

class PyTypesPathStore(PathStore):
    '''
    Path store implementation that stores path information in data types based
    on built-in Python types (dict, list, etc).

    >>> path_store = PyTypesPathStore()

    ''' + PathStore.__doc__

    lock = None

    def __init__(self):
        self.lock = RLock()
        super(PyTypesPathStore, self).__init__()
        self.path_mapping = DictStackWithReverseLookupsWithExplicitKeys()
        self.entries = DictListWithExplicitKeys()
        self.entries.define('/')

    @nonatomic(attr_named('lock'))
    def put_file(self, fake_path, real_path):
        self.put_file_only(fake_path, real_path)
        self._add_entries_for_file_path(fake_path)

    @nonatomic(attr_named('lock'))
    def put_file_only(self, fake_path, real_path):
        try:
            self.path_mapping[fake_path].push(real_path)
        except KeyError:
            self.path_mapping.define(fake_path)
            self.path_mapping[fake_path].push(real_path)

    @nonatomic(attr_named('lock'))
    def _add_entries_for_file_path(self, fake_path):
        while True:
            fake_path, entry = os.path.split(fake_path)

            if (fake_path != '/') and self.is_empty_dir(fake_path):
                self.remove_directory(fake_path)

            if not entry:
                break
            self.add_entry(fake_path, entry)

    @nonatomic(attr_named('lock'))
    def add_entry(self, fake_path, entry):
        try:
            self.entries[fake_path].append(entry)
        except KeyError:
            self.entries.define(fake_path)
            self.entries[fake_path].append(entry)

    @nonatomic(attr_named('lock'))
    def delete_file_only(self, fake_path):
        self.path_mapping[fake_path].pop()
        if not self.path_mapping[fake_path]:
            self.path_mapping.undefine(fake_path)

    @nonatomic(attr_named('lock'))
    def delete_entry(self, fake_path, entry):
        # Remove item matching entry that is closest to end:
        for index, item in reversed(list(enumerate(self.entries[fake_path]))):
            if item == entry:
                self.entries[fake_path].pop(index)
                break

        if not self.entries[fake_path]:
            self.entries.undefine(fake_path)

    @nonatomic(attr_named('lock'))
    def rename_file(self, old_fake_path, new_fake_path):
        if old_fake_path == new_fake_path:
            raise AssertionError(
              'PyTypesPathStore.rename_file: old_fake_path == new_fake_path')

        real_path = self.get_real_path(old_fake_path)

        old_dir = os.path.dirname(old_fake_path)
        new_dir = os.path.dirname(new_fake_path)

        if old_dir == new_dir:
            old_entry = os.path.basename(old_fake_path)
            new_entry = os.path.basename(new_fake_path)

            for index, entry in reversed(
              list(enumerate(self.entries[old_dir]))):
                if entry == old_entry:
                    self.entries[old_dir][index] = new_entry
                    break

            self.delete_file_only(old_fake_path)
            self.put_file_only(new_fake_path, real_path)
        else:
            self.delete_file(old_fake_path)
            self.put_file(new_fake_path, real_path)

    @nonatomic(attr_named('lock'))
    def create_directory(self, fake_path):
        try:
            self.entries.define(fake_path)
        except KeyError:
            raise PathExists(fake_path)
        self._add_entries_for_file_path(fake_path)

    @nonatomic(attr_named('lock'))
    def remove_directory(self, fake_path):
        if fake_path == '/':
            raise ValueError('fake_path must not be "/"')

        try:
            path_entries = self.entries[fake_path]
        except KeyError:
            raise FakePathNotFound(fake_path)

        if path_entries:
            raise DirectoryNotEmpty(fake_path)

        self.entries.undefine(fake_path)
        self._delete_entries_for_file_path(fake_path)

    @nonatomic(attr_named('lock'))
    def delete_file(self, fake_path):
        self.delete_file_only(fake_path)
        self._delete_entries_for_file_path(fake_path)

    @nonatomic(attr_named('lock'))
    def _delete_entries_for_file_path(self, fake_path):
        while True:
            fake_path, entry = os.path.split(fake_path)
            if not entry:
                break
            self.delete_entry(fake_path, entry)

    @nonatomic(attr_named('lock'))
    def get_real_path(self, fake_path):
        try:
            return self.path_mapping.lookup(fake_path)
        except KeyError:
            raise FakePathNotFound(fake_path)

    @nonatomic(attr_named('lock'))
    def get_fake_path(self, real_path):
        try:
            return self.path_mapping.reverse_lookup(real_path)
        except KeyError:
            raise RealPathNotFound(real_path)

    @nonatomic(attr_named('lock'))
    def get_entries_for_directory(self, fake_path):
        return last_unique(self.entries[fake_path])

    def supports_threads(self):
        return True

    @nonatomic(attr_named('lock'))
    def get_entries(self, fake_path):
        if not self.path_exists(fake_path):
            raise FakePathNotFound(fake_path)
        if not self.is_dir(fake_path):
            raise NotADirectory(fake_path)
        entries = self.get_entries_for_directory(fake_path)
        return entries

    @nonatomic(attr_named('lock'))
    def is_file(self, fake_path):
        try:
            return bool(self.get_real_path(fake_path))
        except PathNotFound:
            return False

    @nonatomic(attr_named('lock'))
    def is_dir(self, fake_path):
        return (fake_path in self.entries)

    @nonatomic(attr_named('lock'))
    def is_empty_dir(self, fake_path):
        return self.is_dir(fake_path) and (
          not self.get_entries_for_directory(fake_path))

    @nonatomic(attr_named('lock'))
    def path_exists(self, fake_path):
        if self.is_file(fake_path):
            return True
        if self.is_dir(fake_path):
            return True
        return False
