# Copyright (c) 2007-2008 Forest Bond.
# This file is part of the pytagsfs software package.
#
# pytagsfs is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import sys, os, signal, re, stat
from itertools import chain
from threading import RLock

from pytagsfs.exceptions import (
  PathNotFound,
  PathExists,
  UnrepresentableRealPath,
)
from pytagsfs.util import nonatomic, attr_named

from pytagsfs.debug import (
  log_debug,
  log_info,
  log_warning,
  log_error,
  log_traceback,
  exceptionally_logged,
  verbosely_logged,
)

def _make_filter(expr, real):
    if real:
        get_arg = lambda args: args[0]
    else:
        get_arg = lambda args: args[1]

    if expr.startswith('!'):
        trans_result = lambda x: not x
        regex = re.compile(expr[1:])
    else:
        trans_result = lambda x: x
        regex = re.compile(expr)
        
    def filtr(*args):
        return trans_result(regex.search(get_arg(args)))

    return filtr

STAT = object()
ENTRIES = object()

class SourceTreeRepresentation(object):
    lock = None

    path_parser = None
    path_store = None
    source_tree = None
    monitor = None
    cache = None

    filters = None

    def __init__(
          self, path_parser, path_store, source_tree, monitor,
          cache = None, filters = None, debug = False
        ):
        self.lock = RLock()

        self.debug = debug

        self.path_parser = path_parser
        self.path_store = path_store
        self.source_tree = source_tree
        self.monitor = monitor
        self.cache = cache

        self.filters = []

        if filters:
            for expr, real in filters:
                self.add_filter(expr, real)

    def start(self):
        self.monitor.start(self)
        self.populate()

    def stop(self):
        self.monitor.stop()

    def populate(self):
        log_info('populating source tree representation...')
        self.add_source_dir(self.source_tree.root)

################################################################################

    def parse_path(self, real_path):
        try:
            fake_path = self.path_parser.parse(real_path)
        except UnrepresentableRealPath, e:
            try:
                log_debug(u'Unrepresentable file: %s' % e.path)
            except (UnicodeEncodeError, UnicodeDecodeError):
                log_traceback()
                log_error(
                  u'Unrepresentable file: [failed to convert path to unicode]')
            raise e
        return fake_path

    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('lock'))
    def add_source_dir(self, real_path):
        '''
        Recursively add source directory ``real_path`` to the source tree
        representation.  Fail silently if:

         * Path corresponds with a file that is not a directory.
         * Directory does not exist.
         * os.path.basename(real_path).startswith('.pytagsfs')

        If a directory with the same real path already exists in the source
        tree representation, it will be silently removed from the
        representation before the add operation is executed.
        '''
        if os.path.basename(real_path).startswith('.pytagsfs.'):
            return

        try:
            self.remove_source_dir(real_path)
        except PathNotFound:
            pass

        self.monitor.add_source_dir(real_path)
        for dirpath, dirnames, filenames in self.source_tree.walk(real_path):
            if dirpath != real_path:
                break
            for dirname in dirnames:
                real_path = os.path.join(dirpath, dirname)
                self.add_source_dir(real_path)
            for filename in filenames:
                real_path = os.path.join(dirpath, filename)
                self.add_source_file(real_path)

    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('lock'))
    def remove_source_dir(self, real_path):
        '''
        Recursively remove source directory ``real_path`` from the source tree
        representation.  Do nothing if:

         * ``real_path`` was never added to the source tree representation.
         * os.path.basename(real_path).startswith('.pytagsfs')
        '''
        if os.path.basename(real_path).startswith('.pytagsfs.'):
            return

        try:
            fake_path = self.get_fake_path(real_path)
        except PathNotFound, e:
            log_debug(
              (u'remove_source_dir: trying to remove source dir %s, '
               u'got PathNotFound (already removed?).'
              ) % real_path
            )
            return
        self.cache_prune_branch_to(fake_path)
        self.monitor.remove_source_dir(real_path)

    def update_source_dir(self, real_path):
        pass

    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('lock'))
    def add_source_file(self, real_path):
        '''
        Add source file ``real_path`` to the source tree representation.  Do
        nothing if:

         * The file does not exist.
         * The target file is a directory.
         * os.path.basename(real_path).startswith('.pytagsfs')
        '''
        if os.path.basename(real_path).startswith('.pytagsfs.'):
            return

        if not self.source_tree.isreadable(real_path):
            log_debug(
              u'add_source_file: not readable, not adding: %s' % real_path)
            return

        try:
            self.remove_source_file(real_path)
        except PathNotFound:
            pass

        try:
            fake_path = self.parse_path(real_path)
        except UnrepresentableRealPath:
            return

        if not self.source_tree.isreadable(real_path):
            log_debug(
              u'add_source_file: not readable, not adding: %s' % real_path)
            return

        if not self.filter_path(
          self.source_tree.get_relative_path(real_path),
          fake_path
        ):
            log_debug(u'Explicitly excluded fake path: %s; real path: %s' % (
              fake_path, real_path))
            return

        self.monitor.add_source_file(real_path)

        try:
            self.get_fake_path(real_path)
        except PathNotFound:
            pass
        else:
            log_error(
              'add_source_file: tried to add the same real_path twice: %s' %
              real_path
            )
            return

        self.path_store.put_file(fake_path, real_path)
        self.cache_prune_branch_to(fake_path)

    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('lock'))
    def remove_source_file(self, real_path):
        if os.path.basename(real_path).startswith('.pytagsfs.'):
            return

        try:
            fake_path = self.get_fake_path(real_path)
        except PathNotFound:
            log_debug(u'remove_source_file: PathNotFound: %s' % real_path)
            return
        self.cache_prune_branch_to(fake_path)
        try:
            self.path_store.delete_file(fake_path)
        except PathNotFound:
            log_traceback()
        self.monitor.remove_source_file(real_path)

    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('lock'))
    def update_source_file(self, real_path):
        if os.path.basename(real_path).startswith('.pytagsfs.'):
            return

        try:
            old_fake_path = self.get_fake_path(real_path)
        except PathNotFound:
            log_debug(
              (u'update_source_file: real path not found, '
               u'adding instead: %s'
              ) % real_path
            )
            self.add_source_file(real_path)
            return

        try:
            new_fake_path = self.parse_path(real_path)
        except UnrepresentableRealPath:
            try:
                self.path_store.delete_file(fake_path)
            except PathNotFound:
                log_traceback()
            self.monitor.remove_source_file(real_path)
            return

        if old_fake_path != new_fake_path:
            self.path_store.rename_file(old_fake_path, new_fake_path)

        self.cache_prune_branch_to(old_fake_path)
        self.cache_prune_branch_to(new_fake_path)

    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('lock'))
    def add_source_path(self, real_path):
        self.add_source_file(real_path)
        self.add_source_dir(real_path)

    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('lock'))
    def remove_source_path(self, real_path):
        try:
            self.get_fake_path(real_path)
        except PathNotFound:
            log_debug(u'remove_source_path: PathNotFound: %s' % real_path)
            return
        self.remove_source_file(real_path)
        self.remove_source_dir(real_path)

    def update_source_path(self, real_path):
        self.update_source_file(real_path)

################################################################################

    def add_filter(self, expr, real):
        self.filters.append(_make_filter(expr, real))

    def filter_path(self, real_path, fake_path):
        for filtr in self.filters:
            if not filtr(real_path, fake_path):
                return False
        return True

################################################################################

    def put_file(self, real_path):
        return self.add_source_file(real_path)

    def delete_file(self, fake_path):
        return self.remove_source_file(fake_path)

    @nonatomic(attr_named('lock'))
    def rename_path(self, old_fake_path, new_fake_path):
        log_info(u'SourceTreeRepresentation.rename_path: renaming %s to %s' % (
          old_fake_path, new_fake_path
        ))

        if self.is_dir(old_fake_path):
            entries = self.get_entries(old_fake_path)
            log_debug(
              u'SourceTreeRepresentation.rename_path: entries = %s' %
              repr(entries)
            )
            if not entries:
                self.path_store.remove_directory(old_fake_path)
                self.path_store.create_directory(new_fake_path)
            else:
                for entry in entries:
                    old_entry_path = os.path.join(old_fake_path, entry)
                    new_entry_path = os.path.join(new_fake_path, entry)
                    real_path = self.get_real_path(old_entry_path)
                    self.path_parser.apply(real_path, new_entry_path)
        else:
            real_path = self.get_real_path(old_fake_path)
            self.path_parser.apply(real_path, new_fake_path)

        self.cache_prune_branch_to(old_fake_path)
        self.cache_prune_branch_to(new_fake_path)

    @nonatomic(attr_named('lock'))
    def create_directory(self, fake_path):
        self.path_store.create_directory(fake_path)
        self.cache_prune_branch_to(fake_path)

    @nonatomic(attr_named('lock'))
    def remove_directory(self, fake_path):
        self.path_store.remove_directory(fake_path)
        self.cache_prune_branch_to(fake_path)

    @nonatomic(attr_named('lock'))
    def get_real_path(self, fake_path):
        return self.path_store.get_real_path(fake_path)

    @nonatomic(attr_named('lock'))
    def get_fake_path(self, real_path):
        return self.path_store.get_fake_path(real_path)

    def get_entries(self, fake_path):
        try:
            return self.cache_get(fake_path, ENTRIES)
        except KeyError:
            pass
        self.lock.acquire()
        try:
            entries = self.path_store.get_entries(fake_path)
            self.cache_put(fake_path, ENTRIES, entries)
            return entries
        finally:
            self.lock.release()

    @nonatomic(attr_named('lock'))
    def path_exists(self, fake_path):
        return self.path_store.path_exists(fake_path)

    @nonatomic(attr_named('lock'))
    def is_file(self, fake_path):
        return self.path_store.is_file(fake_path)

    @nonatomic(attr_named('lock'))
    def is_dir(self, fake_path):
        return self.path_store.is_dir(fake_path)

    def supports_threads(self):
        return (
          self.path_store.supports_threads() and (
            self.monitor is None or
            self.monitor.supports_threads()
          )
        )

    def getattr(self, fake_path):
        try:
            return self.cache_get(fake_path, STAT)
        except KeyError:
            pass
        self.lock.acquire()
        try:
            stat_result = self._getattr(fake_path)
            self.cache_put(fake_path, STAT, stat_result)
            return stat_result
        finally:
            self.lock.release()

    def _getattr(self, fake_path):
        try:
            real_path = self.get_real_path(fake_path)
        except PathNotFound:
            pass
        else:
            return os.lstat(real_path)

        # Files should've been handled above.  Now we're just dealing with
        # directories and non-existent paths.

        # If fake_path doesn't exist, the exception will be raised here and
        # caught be our caller.
        subdirs = self.get_subdirectories(fake_path)

        source_root_statinfo = os.lstat(self.source_tree.root)

        st_mode = stat.S_IFDIR | (source_root_statinfo.st_mode & 0777)

        try:
            st_dev = source_root_statinfo.st_dev
            st_uid = source_root_statinfo.st_uid
            st_gid = source_root_statinfo.st_gid
        except AttributeError:
            # source_root_statinfo is actually an integer indicating an error
            log_error(
              u'SourceTreeRepresentation._getattr: source_root_statinfo = %s' %
              repr(source_root_statinfo)
            )
            return source_root_statinfo

        st_ino = 0
        st_nlink = 2 + len(subdirs)
        st_size = 0
        st_atime = self.get_directory_atime(fake_path)
        st_mtime = self.get_directory_mtime(fake_path)
        st_ctime = self.get_directory_ctime(fake_path)

        return os.stat_result((
          st_mode, st_ino, st_dev, st_nlink, st_uid,
          st_gid, st_size, st_atime, st_mtime, st_ctime,
        ))

    @nonatomic(attr_named('lock'))
    def utime(self, fake_path, times):
        real_path = self.get_real_path(fake_path)
        os.utime(real_path, times)
        self.cache_prune_branch_to(fake_path, STAT)

    @nonatomic(attr_named('lock'))
    def get_subdirectories(self, fake_path):
        entries = self.get_entries(fake_path)
        subdirs = [ ]

        for entry in entries:
            entry_path = os.path.join(fake_path, entry)
            try:
                if self.is_dir(entry_path):
                    subdirs.append(entry_path)
            except PathNotFound:
                log_traceback()
                log_error(
                  u'PathNotFound calling is_dir on previously known entry')
                pass
        return subdirs

    def _get_file_time_attr(self, fake_path, attr_name):
        return getattr(self.getattr(fake_path), attr_name)

    @nonatomic(attr_named('lock'))
    def _get_directory_time_attr(self, fake_path, attr_name):
        entries = self.get_entries(fake_path)
        entry_paths = [os.path.join(fake_path, e) for e in entries]

        subdirs = [
          entry_path for entry_path in entry_paths if self.is_dir(entry_path)
        ]
        files = [
          entry_path for entry_path in entry_paths if self.is_file(entry_path)
        ]

        subdir_times = [
          self._get_directory_time_attr(d, attr_name) for d in subdirs ]
        file_times = [
          self._get_file_time_attr(f, attr_name) for f in files ]

        times = list(chain(subdir_times, file_times))

        if not times:
            return 0

        return max(times)

    def get_file_atime(self, fake_path):
        return self._get_file_time_attr(fake_path, 'st_atime')

    def get_directory_atime(self, fake_path):
        return self._get_directory_time_attr(fake_path, 'st_atime')

    def get_file_mtime(self, fake_path):
        return self._get_file_time_attr(fake_path, 'st_mtime')

    def get_directory_mtime(self, fake_path):
        return self._get_directory_time_attr(fake_path, 'st_mtime')

    def get_file_ctime(self, fake_path):
        return self._get_file_time_attr(fake_path, 'st_ctime')

    def get_directory_ctime(self, fake_path):
        return self._get_directory_time_attr(fake_path, 'st_ctime')

################################################################################

    @nonatomic(attr_named('lock'))
    def cache_put(self, fake_path, key, value):
        if self.cache:
            return self.cache.put(fake_path, key, value)

    @nonatomic(attr_named('lock'))
    def cache_get(self, fake_path, key):
        if self.cache:
            return self.cache.get(fake_path, key)
        raise KeyError('no cache to get from')

    @nonatomic(attr_named('lock'))
    def cache_prune(self, fake_path = None, key = None):
        if self.cache:
            return self.cache.prune(fake_path, key)
        raise KeyError('no cache to prune')

    @nonatomic(attr_named('lock'))
    def cache_prune_branch_to(self, fake_path, key = None):
        while True:
            try:
                self.cache_prune(fake_path, key)
            except KeyError:
                pass
            if fake_path == '/':
                break
            fake_path = os.path.dirname(fake_path)
