# Copyright (c) 2007-2008 Forest Bond.
# This file is part of the pytagsfs software package.
#
# pytagsfs is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import re, copy

from pytagsfs.exceptions import ErrorWithMessage

class PatternError(ErrorWithMessage):
    pass

class SubstitutionPattern(object):
    '''
    SubstitutionPattern:

    >>> p = SubstitutionPattern('%a %b')

    >>> d = p.split('foo bar')
    >>> print [ (k, d[k]) for k in sorted(d.keys()) ]
    [('a', 'foo'), ('b', 'bar')]

    >>> p.split('foo')
    Traceback (most recent call last):
    ...
    PatternError: Failed to split string "foo"

    >>> p.fill({'a': 'abc', 'b': 'xyz'})
    'abc xyz'

    >>> d = p.split('foo   bar')
    >>> print [ (k, d[k]) for k in sorted(d.keys()) ]
    [('a', 'foo'), ('b', '  bar')]


    >>> p = SubstitutionPattern('%x %y %x')

    >>> d = p.split('a b c')
    Traceback (most recent call last):
    ...
    PatternError: Got differing values for format key x: a, c

    >>> d = p.split('a b a')
    >>> print [ (k, d[k]) for k in sorted(d.keys()) ]
    [('x', 'a'), ('y', 'b')]

    >>> p.fill({'y': 'foo'})
    Traceback (most recent call last):
    ...
    KeyError: 'x'
    '''
    _format_string = None
    _format_keys = None
    _regex_group_order = None
    _interp_string = None
    _regex_string = None

    def __init__(self, format_string, regex_atom = r'.+?'):
        self._format_string = format_string
        self.regex_atom = regex_atom
        self._find_format_keys()
        self._find_regex_keys()
        self._build_regex()
        self._build_interp_string()

    def __str__(self):
        return self._format_string

    def _find_format_keys(self):
        self._format_keys = re.findall(r'%([^%])', self._format_string)

    def _find_regex_keys(self):
        self._regex_keys = self._format_keys

    def _build_regex(self):
        self._regex_group_order = [ ]

        regex = copy.copy(self._format_string)
        regex = re.escape(regex)

        regex = self._pre_process_regex(regex)

        for k in self._regex_keys:
            regex = regex.replace(
              re.escape('%%%s' % k),
              r'(%s)' % self.regex_atom,
              1
            )
            self._regex_group_order.append(k)

        regex = self._post_process_regex(regex)

        regex = r'^%s$' % regex
        self._regex_string = regex
        self.regex = re.compile(regex)

    def _pre_process_regex(self, regex):
        return regex

    def _post_process_regex(self, regex):
        return regex

    def _pre_process_interp_string(self, interp_string):
        return interp_string

    def _post_process_interp_string(self, interp_string):
        return interp_string

    def _build_interp_string(self):
        interp_string = copy.copy(self._format_string)

        interp_string = self._pre_process_interp_string(interp_string)

        for k in self._format_keys:
            interp_string = interp_string.replace('%%%s' % k, '%%(%s)s' % k, 1)

        interp_string = self._post_process_interp_string(interp_string)

        self._interp_string = interp_string

    def split(self, s):
        result = { }

        mo = self.regex.match(s)
        if mo is None:
            raise PatternError('Failed to split string "%s"' % s)

        groups = mo.groups()
        for k, v in zip(self._regex_group_order, groups):
            if (k in result) and (result[k] != v):
                raise PatternError(
                  'Got differing values for format key %s: %s, %s' % (
                    k, result[k], v)
                )
            result[k] = v

        return result

    def fill(self, parts):
        return self._interp_string % parts

class Section(object):
    def __init__(self, content):
        self.content = content

    def __str__(self):
        return '%s: %s' % (self.__class__.__name__, self.content)

    def evaluate(self, parts):
        return self.content % parts

class ConditionalSection(Section):
    def evaluate(self, parts):
        try:
            return self.content % parts
        except KeyError:
            return ''

class ConditionalSubstitutionPattern(SubstitutionPattern):
    __doc__ = SubstitutionPattern.__doc__.replace(
      'SubstitutionPattern', 'ConditionalSubstitutionPattern'
    ) + '''

    ConditionalSubstitutionPattern:

    >>> p = ConditionalSubstitutionPattern('%?%a%? %b')

    >>> d = p.split('foo bar')
    >>> print [ (k, d[k]) for k in sorted(d.keys()) ]
    [('a', 'foo'), ('b', 'bar')]

    >>> d = p.split(' bar')
    >>> print [ (k, d[k]) for k in sorted(d.keys()) ]
    [('a', None), ('b', 'bar')]

    >>> d = p.split('foo ')
    Traceback (most recent call last):
    ...
    PatternError: Failed to split string "foo "

    >>> p.fill({'a': 'abc', 'b': 'xyz'})
    'abc xyz'

    >>> p.fill({'b': 'xyz'})
    ' xyz'

    >>> p.fill({'a': 'abc'})
    Traceback (most recent call last):
    ...
    KeyError: 'b'

    >>> p = ConditionalSubstitutionPattern('%x%? [%y]%?')

    >>> d = p.split('foo [bar]')
    >>> print [ (k, d[k]) for k in sorted(d.keys()) ]
    [('x', 'foo'), ('y', 'bar')]

    >>> p.fill({'x': 'foo', 'y': 'bar'})
    'foo [bar]'

    >>> p.fill({'x': 'foo'})
    'foo'
    '''

    def _find_format_keys(self):
        self._format_keys = re.findall(r'%([^%?])', self._format_string)

    def _pre_process_regex(self, regex):
        # Turn %?xyz%? into (?:xyz)?
        regex = re.sub(
          ''.join((
            re.escape(re.escape(r'%?')),
            r'(.*?)',
            re.escape(re.escape(r'%?')),
          )),
          r'(?:\1)?',
          regex
        )
        return regex

    def _make_conditional_section(self, s):
        return ConditionalSection(s)

    def fill(self, parts):
        count = 0
        sections = [ ]
        for s in self._interp_string.split('%?'):
            if (count % 2) == 0:
                section = Section(s)
            else:
                section = self._make_conditional_section(s)
            sections.append(section)
            count = count + 1
        evaluated_sections = [ s.evaluate(parts) for s in sections ]
        return ''.join(evaluated_sections)

class IfElseSection(Section):
    default = None

    def __init__(self, content):
        try:
            self.content, self.default = content.split('%:')
        except (ValueError, TypeError):
            self.content = content
            self.default = ''

    def evaluate(self, parts):
        try:
            return self.content % parts
        except KeyError:
            return self.default

class IfElseSubstitutionPattern(ConditionalSubstitutionPattern):
    __doc__ = ConditionalSubstitutionPattern.__doc__.replace(
      'ConditionalSubstitutionPattern', 'IfElseSubstitutionPattern'
    ) + '''

    IfElseSubstitutionPattern:

    >>> p = IfElseSubstitutionPattern('%a %?%b%:unset%?')

    >>> p.fill({'a': 'abc', 'b': 'xyz'})
    'abc xyz'

    >>> p.fill({'a': 'abc'})
    'abc unset'

    >>> p.fill({'b': 'xyz'})
    Traceback (most recent call last):
    ...
    KeyError: 'a'

    >>> d = p.split('foo bar')
    >>> print [ (k, d[k]) for k in sorted(d.keys()) ]
    [('a', 'foo'), ('b', 'bar')]

    >>> d = p.split('foo unset')
    >>> print [ (k, d[k]) for k in sorted(d.keys()) ]
    [('a', 'foo'), ('b', 'unset')]

    '''

    def _find_format_keys(self):
        self._format_keys = re.findall(r'%([^%?:])', self._format_string)

    def _pre_process_regex(self, regex):
        # Turn %?abc%:xyz%? into (?:abc|xyz)
        regex = re.sub(
          ''.join((
            re.escape(re.escape(r'%?')),
            r'(.*?)',
            re.escape(re.escape(r'%:')),
            r'(.*?)',
            re.escape(re.escape(r'%?')),
          )),
          r'(?:\1|\2)',
          regex
        )
        return ConditionalSubstitutionPattern._pre_process_regex(self, regex)

    def _make_conditional_section(self, s):
        if '%:' in s:
            return IfElseSection(s)
        return super(IfElseSubstitutionPattern, self
        )._make_conditional_section(s)

class ModifiedSubstitutionPattern(SubstitutionPattern):
    __doc__ = SubstitutionPattern.__doc__.replace(
      'SubstitutionPattern', 'ModifiedSubstitutionPattern'
    ) + '''

    ModifiedSubstitutionPattern:

    >>> p = ModifiedSubstitutionPattern('%^a %_b %!c')
    >>> d = p.split('foo bar baz')
    >>> print [ (k, d[k]) for k in sorted(d.keys()) ]
    [('a', 'foo'), ('b', 'bar'), ('c', 'baz')]

    >>> p.fill({'a': 'fOO', 'b': 'bAr', 'c': 'BAZ'})
    'FOO bar Baz'
    '''

    MODIFIERS = {
      '^': lambda s: s.upper(),
      '_': lambda s: s.lower(),
      '!': lambda s: s.title(),
    }

    def _find_format_keys(self):
        self._format_keys = [ ]
        self._modifiers = [ ]
        for mk in re.findall(r'%([\^_!]?[^%])', self._format_string):
            if len(mk) > 1:
                m, k = mk[0], mk
            else:
                m, k = None, mk
            self._format_keys.append(k)
            self._modifiers.append(m)

    def _find_regex_keys(self):
        self._regex_keys = [ ]
        for k in self._format_keys:
            self._regex_keys.append(k[-1])

    def _pre_process_regex(self, regex):
        for m in self._modifiers:
            if m:
                regex = regex.replace(
                  re.escape('%%%s' % m),
                  re.escape('%')
                )
        return super(
          ModifiedSubstitutionPattern, self)._pre_process_regex(regex)

    def fill(self, parts):
        parts = dict(parts)
        for m, k in zip(self._modifiers, self._format_keys):
            if m:
                parts[k] = self.MODIFIERS[m](parts[k[1]])
        return super(ModifiedSubstitutionPattern, self).fill(parts)

class ModifiedIfElseSubstitutionPattern(
  ModifiedSubstitutionPattern, IfElseSubstitutionPattern):
    __doc__ = IfElseSubstitutionPattern.__doc__.replace(
      'IfElseSubstitutionPattern', 'ModifiedIfElseSubstitutionPattern'
    ) + ModifiedSubstitutionPattern.__doc__.replace(
      'ModifiedSubstitutionPattern', 'ModifiedIfElseSubstitutionPattern'
    ) + '''

    ModifiedIfElseSubstitutionPattern:

    >>> p = ModifiedIfElseSubstitutionPattern('%^a %!b%? %!c%?')

    >>> p.fill({'a': 'foo', 'b': 'bar'})
    'FOO Bar'

    >>> p.fill({'a': 'foo', 'b': 'bar', 'c': 'BAZ'})
    'FOO Bar Baz'
    '''

    def _find_format_keys(self):
        self._format_keys = [ ]
        self._modifiers = [ ]
        for mk in re.findall(r'%([\^_!]?[^%?:])', self._format_string):
            if len(mk) > 1:
                m, k = mk[0], mk
            else:
                m, k = None, mk
            self._format_keys.append(k)
            self._modifiers.append(m)

    def fill(self, parts):
        parts = dict(parts)
        for m, k in zip(self._modifiers, self._format_keys):
            if m:
                try:
                    parts[k] = self.MODIFIERS[m](parts[k[1]])
                except KeyError:
                    pass

        count = 0
        sections = [ ]
        for s in self._interp_string.split('%?'):
            if (count % 2) == 0:
                section = Section(s)
            else:
                section = self._make_conditional_section(s)
            sections.append(section)
            count = count + 1
        evaluated_sections = [ s.evaluate(parts) for s in sections ]
        return ''.join(evaluated_sections)
