# Copyright (c) 2007-2008 Forest Bond.
# This file is part of the pytagsfs software package.
#
# pytagsfs is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

def now():
    import time
    return int(time.time())

def unique(iter):
    '''
    Generator; iterates over iter and yields each item that has not been seen
    already.  This algorithm prefers items toward the beginning of the list.

    >>> list(unique([1, 2, 5, 4, 5, 2, 3, 6, 2]))
    [1, 2, 5, 4, 3, 6]
    '''
    seen = []
    for item in iter:
        if item not in seen:
            yield item
        seen.append(item)

def last_unique(iter):
    '''
    Like unique, but prefer items toward the end of the list.  Also, returns
    a list rather than an iterator object.

    Given a list l, last_unique(l) is functionally equivalent to
    list(reversed(list(unique(reversed(l))))).

    >>> last_unique([1, 2, 5, 4, 5, 2, 3, 6, 2])
    [1, 4, 5, 3, 6, 2]
    '''
    result = []
    for item in iter:
        try:
            result.remove(item)
        except ValueError:
            pass
        result.append(item)
    return result

def nonatomic(get_lock):
    try:
        from functools import wraps
    except ImportError:
        from sclapp.legacy_support import wraps

    def decorator(wrapped):
        @wraps(wrapped)
        def fn(*args, **kwargs):
            if hasattr(get_lock, '__call__'):
                lock = get_lock(*args, **kwargs)
            else:
                lock = get_lock
            lock.acquire()
            try:
                return wrapped(*args, **kwargs)
            finally:
                lock.release()
        return fn
    return decorator

def attr_named(name):
    def get_attr_by_name(self, *args, **kwargs):
        return getattr(self, name)
    return get_attr_by_name

def rpartition(s, by):
    '''
    In Python 2.4, str's have no method "rpartition".  This function tries to
    use Python 2.5's str method, and falls back to a custom implementation.
    '''
    try:
        return s.rpartition(by)
    except AttributeError:
        pass
    parts = s.split(by)
    if len(parts) == 1:
        return ('', '', s)
    return (by.join(parts[:-1]), by, parts[-1])

def get_obj_by_dotted_name(dotted_name):
    from sclapp.util import importName

    modname, dot, objname = rpartition(dotted_name, '.')
    mod = importName(modname)
    return getattr(mod, objname)

def SimpleMutagenFile(filename):
    import mutagen
    from mutagen.id3 import ID3FileType
    from mutagen.easyid3 import EasyID3

    f = mutagen.File(filename)
    if isinstance(f, ID3FileType):
        f = f.__class__(filename, ID3 = EasyID3)

        # Odd that this is necessary.  Is it a mutagen bug?
        # http://www.listen-project.org/ticket/714
        # http://www.listen-project.org/changeset/841
        if f.tags is None:
            f.tags = EasyID3()
    return f

def stat_result_to_fuse_stat(stat_result):
    import fuse
    return fuse.Stat(
      st_mode = stat_result.st_mode,
      st_ino = stat_result.st_ino,
      st_dev = stat_result.st_dev,
      st_nlink = stat_result.st_nlink,
      st_uid = stat_result.st_uid,
      st_gid = stat_result.st_gid,
      st_size = stat_result.st_size,
      st_atime = stat_result.st_atime,
      st_mtime = stat_result.st_mtime,
      st_ctime = stat_result.st_ctime,
    )
